// SearchDocument.java

// SearchToHTML copyright (c) 1999-2001 David Faden. All rights reserved.
//
// The applet class files and source code are distributed as linkware.
// If you use this applet or a variant on its code,
// include a link to The Gilbert Post, 
// http://www.geocities.com/gilbertnews/
//
// The code and binaries are distributed without warranty of any kind.
// The Gilbert Post and David Faden take no responsibility
// for anything bad that happens as a result of using this applet.
// Please send reports of problems to gilbertnews@hotmail.com, anyway, though.
//

// Modification log:
//
// 4/12/2000 fixed a "Y2K bug" reported by several alert users...  I am not sure what 
// I was thinking when I wrote the portion of code calling Date.getYear()... Perhaps that it
// returns the decade? Anyway, in reality, getYear() returns the number of years
// since 1900. Files with modification dates beyond 1999 were listed with dates greater than
// 99 (100 for 2000).
// Note: the whole Date class is deprecated in JDK 1.1
// The code actually changed is found in HDocSearcher.java.
// 
// 4/12/2000 added code that causes the HDocSearcher's runner Thread to wait
// when it is not "doing anything." This should be more efficient than in the
// previous incarnation, where runner would sleep, then periodically wake up to 
// see if there was anything to search.
//
// Jul 12 2000 I added a kludgy method to HDocSearcher which will
// finish extracting the title from a document even if a match is
// found within the title. I had been reminded of this behavior several
// times before, but it was Danny Narayan's complaint that spurred me to action.
// See HDocSearcher.finishTitle(StringBuffer).
// Finishing the title exposed another problem: the boolean inTitle was not
// set to false even on finding the end of a title.
//
// Jul 13 2000 I seem to be writing a lot broken sentences in this bug log.
// But that okay.
// Changed the name of the method "foundNoMatch" to "receiveNoMatch." Again,
// I think that the former name was misleading. Added two new parameters to
// deal with expanded context capabilities: leadingContextLength and
// trailingContextLength - leadingContextLength is very misleadingly named.
// I will probably change it tomorrow. The new parameters I was alluding to
// are "leadingcontextlength" and "trailingcontextlength". Not yet documented! 
// I added a new method to HDocSearcher.java: appendTrailingContext(StringBuffer)
// and changed HDocSearcher's constructor in connection with the new trailing context
// stuff.
//
// Jul 15 2000 Fixed "bugs" in HDocSearcher.java that would cause an 
// ArrayOutOfBoundsException to be thrown if leadingContextLength==0. Previous to a few
// edits ago, I had required that this value be greather than zero so the code's
// assumption had been a safe one.
//
// Jul 25 2000 Fixed a bug in appendTrailingContext(StringBuffer). The fix required that
// the method not append directly from the input stream to the context (this was the source
// of the problem) so I renamed appendTrailingContext(StringBuffer) to getTrailingContext().
//
// Aug 17, 2000 Changed code to look directly for EOF and on EOF, to pass 
// '\0' to the SearchSieves rather than (char)-1. 
//
// August 22, 2000 Spruced up the documentation of a few methods.
//
// March 9, 2001 Changed the names of several variables and methods (all private)
// to better reflect my current favored style -- that is first word lower case, 
// following words with initial caps. Edited closeConnection so that bis will
// always be set to null, even on an IOException on closeConnection. Messed around
// with conditional statements in several places to try to test for more common
// cases first.
//
// March 10, 2001 Renamed a few more variables to better reflect their function
// and the preferred naming style. Major changes to the internals of doSearch()
// were made, for the sake of efficiency and clarity.
//
// March 11, 2001. Continued the work. Added many more comments. SearchToHTML
// no longer uses indeces to refer to HDocSearchers. Rather, the HDocSearchers
// now pass a reference to themselves back to their parent. Removed the setErrored()
// method.
// SearchSieve has been changed to take a String in its constructor's arguments
// rather than a char array.
//
// March 13, 2001. Information about the document being searched is pulled
// from the HTTP headers in openConnection(). I've changed this method to
// add file size and or creation date information only if that information
// appears to be valid. If the file size <= 0 or the year is less than 1950,
// that respective portion of the information is discarded. Also, a comma is
// now displayed between the month-date and the year. Sorry to everyone
// outside of the U.S.
//
// March 18, 2001. I am going to make quick and dirty edits to this file.
// Proceed with caution. HDocSearcher has been renamed to SearchDocument.
// SearchDocument does not report matches back to a parent SearchToHTML.
// 
// March 19, 2001. SearchDocument (based on HDocSearcher) no longer implements
// Runnable. The searchFor method now blocks until a search is complete and
// returns true or false depending on whether a search term was found. One
// Thread may interrupt another Thread's search by calling interruptSearch.
// An InterruptedException will be thrown from searchFor or searchFor will
// exit normally.
//
// Changed SearchDocument to use the new CircularStringBuffer
// in collecting the context of a match. This is following the example of 
// Marco Soldati, who sent an improved version of the SearchToHTML applet
// to me several months ago. Before, SearchDocument (formerly HDocSearcher)
// implemented the equivalent of a CircularStringBuffer loose within the code.
// 
// Fixed a long standing bug: if a match was found in the title of a document
// and trailingContextLength > 0, the context would be collected, advancing the
// the text stream, and then an attempt was made to collect the title. Bad. Bad.
//
// Two methods were removed: finishTitle(StringBuffer), getTrailingContext().
//
// April 4, 2001 I added methods that allow a search to be stopped on
// hitting a pattern or started only after seeing a pattern. See: 
// getStartPattern setStartPattern(String) stopUsingStartPattern
// isUsingStartPattern setStopPattern(String) getStopPattern 
// stopUsingStopPattern isUsingStopPattern
// 
// When I added support for start and stop patterns to the searchFor method,
// I also rearranged the innards of that method. The code for completing
// the context and a partially retrieved title has been moved out of the 
// search loop. A couple of returns and continues were cut out -- I hope
// this will make it easier for people to follow what is going on in the code.
//
// June 28, 2001. Corrected a serious flaw in the SearchDocument class
// by modifying the CircularStringBuffer class. This bug was pointed out by
// Frank Verni. SearchDocument was depending on it being legal to have a zero 
// length buffer. CircularStringBuffer would throw an IllegalArgumentException 
// on being passed zero for a buffer length. The easiest, and perhaps not too
// terrible, solution was to make it possible to have a zero length
// CircularStringBuffer.
//
// July 4, 2001. Happy Independence Day, USA! SearchDocument (and so SearchToHTML)
// now understands Unix-style relative paths using the buildURL method. It seems fairly
// obvious that buildURL belongs in a separate class acting as an pseudoextension
// to URL, but, for now, I'm sticking it in SearchDocument. I should probably
// move some of the utility methods from SearchToHTML to SearchDocument to
// reduce coupling. It was Scott Beach's request that prompted this feature to
// finally be added.
//
// July 6, 2001. Added support for AND searches -- searches that require
// every term to match before the document is considered to match -- in response
// to a request from Simone Lazzarini. Previously, SearchDocument had only supported 
// OR searches. For an OR search, if a single term matches, the document is
// considered to match.
//
// August 21, 2001. Fixed a bug in the implementation of AND searches. Before, under an
// AND search, some terms would be skipped. The fact that this wasn't caught sooner
// is explained by how hard it was formerly to do AND searches.

import java.io.*;
import java.net.*;
import java.util.*;

//import CircularStringBuffer;
//import SearchSieve;

/**
 * This class represents a searchable document.
 *
 * @author David Faden <a href="mailto:gilbertnews@hotmail.com">gilbertnews</a>
 * @version SearchDocument 1.5 August 21, 2001
 */
public class SearchDocument {
  	private static final String[] months = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};

  
    /**
     * Stream of text to search.
     */
    private BufferedInputStream bis;
    //XXX! This should not be a class variable -- Only kept this way
    //to take advantage of HDocSearcher's openConnection and closeConnection
    //methods. When I have more time, I will chop this out.
  
    /**
     * Setting this to true while a search is underway will cause the 
     * searchFor method to throw an InterruptedException.
     */
    private volatile boolean interruptSearch;
      
    /**
     * True if information about the document to be searched has been collected.
     */ 
    private boolean gotDocumentInfo;
    
    /**
     * True if the title of the document has been parsed.
     */
    private boolean gotTitle;
    
    /**
     * If there is a horrible error such as a MalformedURLException
     * or a SecurityException, refuse to search anymore.
     */
    private volatile boolean noHorribleError;
      
    /**
     * The URL of the document to search.
     */
    private URL url;
    
    /**
     * The relative URL given in the constructor.
     */
    private String relativeURL;
    
    /**
     * The title of the document being searched.
     */
    private String title;
    
    /**
     * String containing information about the document being searched.
     */
    private String documentInfo;
  
    /**
     * String holding the context of the last search.
     * It will be null if there was no match or the search was interrupted.
     */
    private String contextString;
    
    /**
     * The last HTML anchor encountered before a match was found.
     */
    private String nearestAnchor;
  
    /**
     * How many characters up to and including the match should be recorded.
     */
    private int leadingContextLength;
  
    /**
     * How many characters following the match should be recorded.
     */
    private int trailingContextLength;
      
    /**
     * Used to find the beginning of the title.
     */
    private SearchSieve titleFinder = new SearchSieve("<title>", false);
    
    /**
     * Used to find the beginning of HTML anchors.
     */
    private SearchSieve anchorFinder = new SearchSieve("<a name=", false);
    
    /**
     * Stop searching and return false as soon as this pattern is seen.
     */
    private String stopPattern = null;
    
    /**
     * Control whether the stop pattern causes a stop.
     */
    private boolean useStopPattern = false;
    
    /**
     * Start searching only after this pattern is seen.
     */
    private String startPattern = null;
    
    /**
     * Control whether the start pattern is needed to start a search.
     */
    private boolean useStartPattern = false;
    
  
    /**
     * Construct a new SearchDocument.
     * <br>
     * This SearchDocument represents the document
     * at the URL formed by taking baseURL and relativeURL together.
     * <br>
     * @param baseURL With relativeURL, gives the address of the document to search.
     * @param relativeURL Relative to baseURL, gives the address of the document to search.
     * @param leadingContextLength Number of characters up to and including match to
     *          be sent with search results as the match context.
     * @param trailingContextLength Number of characters after match to send as the match
     *           context.
     * @throws MalformedURLException If baseURL and relativeURL together don't form a
     *   valid URL.
     * @throws IllegalArgumentException If leadContextLength &lt; 0.
     * @throws IllegalArgumentException If trailingContextLength &lt; 0.
     */
    public SearchDocument(URL baseURL, 
                        String relativeURL,
                        int leadingContextLength,
                        int trailingContextLength) throws MalformedURLException {
        noHorribleError = false;
        url = buildURL(baseURL, relativeURL);
        noHorribleError = true;
        
        gotDocumentInfo = false;
        gotTitle = false;
        title = "";
        documentInfo = "";
        this.relativeURL = relativeURL;
        
        interruptSearch = false;
        
        //Set up for context capture.
        
        contextString = null;
        if (leadingContextLength < 0)
            throw new IllegalArgumentException("leadingContextLength < 0");
        this.leadingContextLength = leadingContextLength;
        
        if (trailingContextLength < 0)
            throw new IllegalArgumentException("trailingContextLength < 0");
        this.trailingContextLength = trailingContextLength;
    }
  
    /**
     * Interpret the relative path in <code>relativePath</code>
     * with respect to the base URL <code>base</code> to create
     * a new URL.
     * <br>
     * This method "properly" handles query strings and Unix style relative paths:
     * <ol>
     *   <li>../ (one directory up)
     *   <li>./ (present directory)
     *   <li>/ (root/top directory)
     *</ol>
     * Note that SearchToHTML will pass the
     * relative path directly onto the Web browser in displaying
     * the search results and that many browsers don't know how
     * to handle "./" "properly."
     *
     * @param base The base URL.
     * @param relativePath String containing a file path relative to base.
     * @return A new URL relative to <code>base</code> according to <code>relativePath</code>.
     * @throws MalformedURLException if relativePath contains enough "../"'s to pop
     * above the root directory level or contains something to confound the URL
     * constructor.
     */
    private static final URL buildURL(URL base, 
                                      String relativePath) 
                                        throws MalformedURLException {
        URL u = null;
        int relPathLength = relativePath.length();
        if (relativePath.startsWith("../")) {
            int relStartIndex = 3; //Index after last ../
            String path = base.getFile();
            
            int dirIndex = path.lastIndexOf('/');
            
            if (dirIndex != -1) {
                dirIndex = path.lastIndexOf('/', dirIndex - 1);
            }
            
            while (relativePath.startsWith("../", relStartIndex)) {
                dirIndex = path.lastIndexOf('/', dirIndex - 1);
                relStartIndex += 3;
                if (dirIndex == -1)
                    throw new MalformedURLException("Higher than root: " + base.toString() +
                                                    " + " + relativePath);
            }
            
            if (relStartIndex != relPathLength)
                relativePath = relativePath.substring(relStartIndex, relPathLength);
            else
                relativePath = "";
                
            //If the relativeURL hasn't climbed all the way to the root
            //directory, cut the base path to match the directory level
            //calculated in the above code.
            //Otherwise, simply set the path equal to the relativePath
            //with a slash appended to the front to indicate that the
            //relativePath has climbed high enough to become the root
            //directory in the new URL.
            if (dirIndex != -1) {
                path = path.substring(0, dirIndex + 1); //Leave the trailing slash
                if (relStartIndex != relPathLength)
                    path += relativePath;
            }
            else
                path = "/" + relativePath;
                
            u = new URL(base.getProtocol(), base.getHost(), path);
        }
        else if (relativePath.startsWith("/")) {
            u = new URL(base.getProtocol(), base.getHost(), 
                                relativePath);
        }
        else if (relativePath.startsWith("./")) {
            u = new URL(base, relativePath.substring(2, relPathLength));
        }
        else if (relativePath.startsWith("?")) {
            u = new URL(base.getProtocol(), base.getHost(), base.getFile() + relativePath);
        }
        else {
            u = new URL(base, relativePath);
        }
        return u;
    }
  
  
    /**
     * Return the title of the document being searched.
     * If the title has not yet been parsed, an empty String is returned.
     *
     * @return String representing the title of the document being searched.
     */
    public String getDocumentTitle() {
        return title;
    }
    
    /**
     * Return information about the document being searched: its file size and 
     * its last modification date.
     *
     * @return String with the file size and last modification date.
     */
    public String getDocumentInfo() {
        return documentInfo;
    }
    
    /**
     * Return the relative URL of the document being searched, as compared
     * to the supplied base URL. This relative URL is what is used in constructing
     * a link in the search results.
     *
     * @return String representing the relative URL.
     */
    public String getRelativeURL() {
        return relativeURL;
    }
    
    public String getNearestAnchor() {
        return nearestAnchor;
    }
  
    /**
     * Returns the last leadingContextLength chars to pass through 
     * the SearchSieves to give more of an idea to the user of what 
     * was around a match.
     * <br>
     * This is not set up to deal with the case that context is still changing as
     * it is asked for or concurrent calls to getContext()...reuses some variables.
     *
     * @return The context of the match.
     */
    public final String getContext() {
		return contextString;
    }
    
    /**
     * Returns true if this HDocSearcher has encountered an error that
     * makes it impossible to continue searching -- an inaccessible document.
     * <br>
     * If isErrored() returns true, the SearchToHTML will ignore
     * all search requests.
     *
     * @return True if this SearchToHTML has an encountered a fatal error.
     */
    public final boolean isErrored() {
        return !noHorribleError;
    }
    
    public void setStartPattern(String pattern) {
        if (pattern == null)
            throw new NullPointerException("null pattern");
            
        this.startPattern = pattern;
        useStartPattern = true;
    }
    
    public String getStartPattern() {
        return startPattern;
    }
    
    public void stopUsingStartPattern() {
        useStartPattern = false;
    }
    
    public boolean isUsingStartPattern() {
        return useStartPattern;
    }
    
    public void setStopPattern(String pattern) {
        if (pattern == null)
            throw new NullPointerException("null pattern");
        this.stopPattern = pattern;
        useStopPattern = true;
    }
    
    public String getStopPattern() {
        return stopPattern;
    }
    
    public void stopUsingStopPattern() {
        useStopPattern = false;
    }
    
    public boolean isUsingStopPattern() {
        return useStopPattern;
    }
    
    /**
     * Search for the patterns contained in <code>searchTerms</code>.
     * <br>
     * If <code>exactMatchesOnly</code> is true, non-word characters
     * must surround any sequence that matches a pattern for that
     * sequence to be considered a match.
     * <br>
     * If <code>ignoreTagContent</code> is true, all text between
     * &lt; and &gt; characters is excluded from the search.
     * <br>
     * If any of the search terms is found, true is returned. Otherwise,
     * false is returned.
     * <br>
     * This method does not block for the search to finish. A search may be
     * canceled by calling <code>stopSearch</code>. If <code>searchFor</code> is
     * called while another search is underway, that other search is cancelled
     * before the new search is started.
     *
     * @param searchTerms An array of patterns.
     * @param exactMatchesOnly No substring matches.
     * @param ignoreTagContent Ignore text in between &lt; and &gt;.
     * @param mustMatchAll Consider a document to match only if it
     * contains a match for every term.
     * @return True if any of the terms matches -- false, otherwise.
     */
    public final synchronized boolean searchFor(String[] searchTerms, 
                                boolean exactMatchesOnly, 
                                boolean ignoreTagContent,
                                boolean mustMatchAll) throws InterruptedException {
                                
        contextString = null;
        nearestAnchor = null;
        
        if (!noHorribleError) 
        	return false; //XXX! Should probably throw Exception here.
        
        boolean foundMatch = false;
        
        int b;
        char c;
        char lowerc;
        
        boolean inTag = false;
        boolean inAnchor = false;
        boolean inTitle = false;
        
        SearchSieve stopFinder = null;
        boolean foundStopPattern = false;
        
        SearchSieve startFinder = null;
        boolean foundStartPattern = true; //Set to false later, if need be.
        
        CircularStringBuffer contextBuffer = null;
        StringBuffer currentAnchorBuffer = new StringBuffer();
        StringBuffer titleBuffer = null; //Satisfy compiler that titleBuffer is initialized. 
        
        anchorFinder.reset();
        
        int numberOfSearchTerms = searchTerms.length;
        SearchSieve[] searchers = new SearchSieve[numberOfSearchTerms];
        
        for (int i = 0; i < numberOfSearchTerms; i++) {
            searchers[i] = new SearchSieve(searchTerms[i].toLowerCase(), exactMatchesOnly);
        }
        
        if (useStopPattern) {
            stopFinder = new SearchSieve(stopPattern, false);
        }
        
        if (useStartPattern) {
            foundStartPattern = false;
            startFinder = new SearchSieve(startPattern, false);
        }
        
        if (!gotTitle) {
            titleFinder.reset();
            titleBuffer = new StringBuffer();
        }
    
        if (!openConnection()) {
            return false; //XXX! Should probably rethrow Exception
        }
        
        if (exactMatchesOnly) {
            contextBuffer = new CircularStringBuffer(leadingContextLength + 1);
        }
        else {
            contextBuffer = new CircularStringBuffer(leadingContextLength);
        }
        
        do {
        
            if (interruptSearch) {
                closeConnection();
                throw new InterruptedException(relativeURL);
            }
        
            try {
                b = bis.read();
                if (b == -1)
                    c = '\0';
                else
                    c = (char) b;
            }
            catch (IOException ioe) {
                ioe.printStackTrace();
                //XXX! Should probably rethrow Exception
                closeConnection(); 
                return false;
            }
       
            lowerc = Character.toLowerCase(c);
            
            if (!inAnchor) 
                inAnchor = anchorFinder.addChar(lowerc);
            else {
                switch (lowerc) {
                    case '\"': case '\'': case ' ': case '>':
                    case '\n': case '\r': case '\t': case '\0':
                        if (currentAnchorBuffer.length() != 0) {
                            nearestAnchor = currentAnchorBuffer.toString();
                            inAnchor = false;
                            currentAnchorBuffer.setLength(0);
                        }
                        break;
                    default:
                        currentAnchorBuffer.append(c);
                }
            
            }
            
            if (!gotTitle) {
                if (!inTitle) 
                    inTitle = titleFinder.addChar(lowerc);
                else {
                    if (lowerc == '<' || b == -1) {
                        if (titleBuffer.length() != 0) {
                            title = titleBuffer.toString(); //class variable
                        }
                        titleBuffer = null;
                        titleFinder = null;
                        gotTitle = true;
                        inTitle = false;
                    }
                    else 
                        titleBuffer.append(c);
                }
            }
       
            //XXX! JavaScripts and comments might louse up this
            //current cheesy method for cutting HTML tags.
            //
            //Look below the search block for the matching
            //conditional that sets inTag to false on seeing
            //a greater than sign character.
            if (lowerc == '<')
                inTag = true;
       
            //Begin search block
            if ((!ignoreTagContent || !inTag) && foundStartPattern) {
            
                contextBuffer.append(c);
         
                for (int i = 0; i < numberOfSearchTerms && !foundMatch; i++) {
                    if (searchers[i].addChar(lowerc)) {
                        if (!mustMatchAll)
                            foundMatch = true;
                        else {
                            numberOfSearchTerms--;
                            if (numberOfSearchTerms == 0)
                                foundMatch = true;
                            else {
                                //If the matching element is not at the end of 
                                //the array, scoot over the rest of the array
                                //to cover it up.
                                //Since numberOfSearchTerms has been decremented,
                                //it is now equal to the last valid index in the array.
                                if (i != numberOfSearchTerms) {
                                    System.arraycopy(searchers, i + 1,
                                        searchers, i, numberOfSearchTerms - i);
                                    i--;
                                    //The terms have been slid to the left so
                                    //the term at i has not been examined.
                                    //Decrement i so that it will be the next time
                                    //round this loop.
                                }
                                //Null out the section of searchers array
                                //that's no longer in user. 
                                searchers[numberOfSearchTerms] = null;
                            }
                        }
                    }
                }
            }
            //End search block
            
            if (lowerc == '>') {
                inTag = false;
            }
            
            if (useStopPattern && !foundStopPattern && foundStartPattern) {
                foundStopPattern = stopFinder.addChar(c);
                if (SearchToHTML.DEBUG) {
                    if (foundStopPattern)
                        System.err.println(relativeURL + ": found stop pattern");
                }
            }
            
            if (useStartPattern && !foundStartPattern) {
                foundStartPattern = startFinder.addChar(c);
                if (SearchToHTML.DEBUG) {
                    if (foundStartPattern)
                        System.err.println(relativeURL + ": found start pattern");
                }
            }
            
            
        } while (b != -1 && !foundMatch && !foundStopPattern);

        //Finish getting a title.
        //Get the trailing context.
        //XXX! Should this be in a separate method?
        if (b != -1 && (foundMatch || inTitle)) {

            int appendedContextLength = 0;
            
            StringBuffer wholeContextBuffer = null;
            
            int lengthToAppend = -1;
            
            if (foundMatch) {
                if (exactMatchesOnly && trailingContextLength > 0) {
                    lengthToAppend = trailingContextLength - 1;
                }
                else {
                    lengthToAppend = trailingContextLength;
                }
            
                wholeContextBuffer = 
                    new StringBuffer(contextBuffer.getSize() + lengthToAppend);
                wholeContextBuffer.append(contextBuffer.toString());
            }
            
            while (inTitle || 
                    (foundMatch && (appendedContextLength < lengthToAppend))) {
                try {
                    b = bis.read();
                }
                catch (IOException ioe1) {
                    ioe1.printStackTrace();
                    //XXX! Should probably rethrow.
                    b = -1;
                }
                
                if (b == -1) { //EOF
                    if (inTitle && !gotTitle) {
                        inTitle = false;
                        gotTitle = true;
                        title = titleBuffer.toString();
                        titleFinder = null;
                    }
                    break;
                }
                
                c = (char) b;
                
                if (b == '<') {
                    inTag = true;
                    if (inTitle && !gotTitle) {
                        inTitle = false;
                        gotTitle = true;
                        title = titleBuffer.toString();
                        titleFinder = null;
                    }
                }
                
                if (b == '>' && inTag) {
                    if (inTitle && !gotTitle) {
                        titleBuffer.append(c);
                    }
                    inTag = false;
                    if (ignoreTagContent)
                        continue;
                }
                
                if (inTitle && !gotTitle)
                    titleBuffer.append(c);
                
                if (foundMatch && (!inTag || !ignoreTagContent) && 
                        (appendedContextLength < lengthToAppend)) {
                    wholeContextBuffer.append(c);
                    appendedContextLength++;
                }
                
            }
            
            if (foundMatch) {
                contextString = wholeContextBuffer.toString();
                contextString = SearchToHTML.makeHTMLSafe(contextString);
            }
        }

        closeConnection();
        return foundMatch;
    }

  
    /**
     * Signal this SearchDocument to abruptly end a search if one is underway.
     * Block until the signal has been acted on.
     */
    public final void interruptSearch() {  
        interruptSearch = true;
        synchronized (this) {
            interruptSearch = false;
        }
    }

    /**
     * Try to open a stream from the URL to be searched.
     * <br>
     * Pull a little information about the file from
     * the HTTP headers.
     * <br>
     * This method catches FileNotFoundException, IOException, and 
     * SecurityException. On receiving a FileNotFoundException or 
     * SecurityException, the SearchToHTML concludes that its URL
     * is unavailable. It then becomes unresponsive to all future
     * search requests. isErrored() will then return true.
     *
     * @return True if the attempt to open the connection was successful 
     * -- otherwise, false.
     */
    private boolean openConnection() {
        try {
            URLConnection uc = url.openConnection();
            if (!gotDocumentInfo) {
                gotDocumentInfo = true;
                Date d = new Date(uc.getLastModified());
                int yearsSince1900 = d.getYear();
                int length = uc.getContentLength();
                StringBuffer documentInfoBuffer = new StringBuffer();
                
                if (length > 1024) {
                    documentInfoBuffer.append(length/1024);
                    documentInfoBuffer.append('k');
                }
                else if (length > 0) {
                    documentInfoBuffer.append(length);
                    documentInfoBuffer.append('b');
                }
                
                if (yearsSince1900 > 71) { //1971
                    documentInfoBuffer.append(", ");
                    documentInfoBuffer.append(months[d.getMonth()]);
                    documentInfoBuffer.append(' ');
                    documentInfoBuffer.append(d.getDate());
                    documentInfoBuffer.append(", ");
                    documentInfoBuffer.append(yearsSince1900 + 1900);
                }
                
                if (documentInfoBuffer.length() > 0) {
                    documentInfo = documentInfoBuffer.toString();
                }
                
            }
            bis = new BufferedInputStream(uc.getInputStream());
        }
        catch (FileNotFoundException fnfe) {
            fnfe.printStackTrace();
            closeConnection();
            noHorribleError = false;
            return false;
        }
        catch (IOException ioe) {
            ioe.printStackTrace();
            closeConnection();
            return false;
        }
        catch (SecurityException se) {
            se.printStackTrace();
            noHorribleError = false;
            closeConnection();
            return false;
        }
  
        return true;
    }
  
    /**
     * Attempt to close BufferedInputStream bis.
     * Regardless of whether the attempt succeeds, set
     * the reference to the dead stream to null.
     */
    private final void closeConnection() {
        try {
            if (bis != null) 
                bis.close();
        }
        catch (IOException ioe) {
            ioe.printStackTrace();
        }
        finally {
            bis = null;
        }
    }
  
}
